# How to implement a new Kubeless run time 

As of today, Kubeless supports the following languages:

* Ruby
* Node.js
* Python

Each language interpreter is implemented as an image of a Docker container, dispatched by the Kubeless controller.

All the structures that implement  language support are coded in the file `k8sutil.go`, located under the `<..>kuberless/pkg/utils` directory of the project tree repository.

If you want to extend it and make another language available (for example .netcore, which is already in progress! :-) it is necessary to change the following components:

## 1. Change the Const block string structure

In this block of constants declaration, you sould create an entry pointing to the repository of the Docker container image for the runtime of the new language.

Usually there are two entries - one for HTTP triggers and another for event based functions. If your new runtime implementation will support only HTTP triggers, then create only one entry as follows:

```
const (
        python27Http   = "bitnami/kubeless-python@sha256:6789266df0c97333f76e23efd58cf9c7efe24fa3e83b5fc826fd5cc317699b55"
        python27Pubsub = "bitnami/kubeless-event-consumer@sha256:5ce469529811acf49c4d20bcd8a675be7aa029b43cf5252a8c9375b170859d83"
        python34Http   = "bitnami/kubeless-python:test@sha256:686cd28cda5fe7bc6db60fa3e8a9a2c57a5eff6a58e66a60179cc1d3fcf1035b"
        python34Pubsub = "bitnami/kubeless-python-event-consumer@sha256:8f92397258836e9c39948814aa5324c29d96ff3624b66dd70fdbad1ce0a1615e"
        node6Http      = "bitnami/kubeless-nodejs@sha256:b3c7cec77f973bf7a48cbbb8ea5069cacbaee7044683a275c6f78fa248de17b4"
        node6Pubsub    = "bitnami/kubeless-nodejs-event-consumer@sha256:b027bfef5f99c3be68772155a1feaf1f771ab9a3c7bb49bef2e939d6b766abec"
        node8Http      = "bitnami/kubeless-nodejs@sha256:1eff2beae6fcc40577ada75624c3e4d3840a854588526cd8616d66f4e889dfe6"
        node8Pubsub    = "bitnami/kubeless-nodejs-event-consumer@sha256:4d005c9c0b462750d9ab7f1305897e7a01143fe869d3b722ed3330560f9c7fb5"
        ruby24Http     = "bitnami/kubeless-ruby@sha256:97b18ac36bb3aa9529231ea565b339ec00d2a5225cf7eb010cd5a6188cf72ab5"
        ruby24Pubsub   = "bitnami/kubeless-ruby-event-consumer@sha256:938a860dbd9b7fb6b4338248a02c92279315c6e42eed0700128b925d3696b606"
        netcore2Http   = "mydocker/kubeless-netcore:latest"
        busybox        = "busybox@sha256:be3c11fdba7cfe299214e46edc642e09514dbb9bbefcd0d3836c05a1e0cd0642"
        pubsubFunc     = "PubSub"
        schedFunc      = "Scheduled"
)
``` 
In the example below, the `netcore2Http` entry was created, pointing to an image of a Docker container running the .netcore runtime.

## 2. Include a runtime variable

Each runtime has a variable for it and needs also to be treated by the function `init()` as shown below.

```
var python, node, ruby, netcore []runtimeVersion

func init() {
        python27 := runtimeVersion{runtimeID: "python", version: "2.7", httpImage: python27Http, pubsubImage: python27Pubsub}
        python34 := runtimeVersion{runtimeID: "python", version: "3.4", httpImage: python34Http, pubsubImage: python34Pubsub}
        python = []runtimeVersion{python27, python34}

        node6 := runtimeVersion{runtimeID: "nodejs", version: "6", httpImage: node6Http, pubsubImage: node6Pubsub}
        node8 := runtimeVersion{runtimeID: "nodejs", version: "8", httpImage: node8Http, pubsubImage: node8Pubsub}
        node = []runtimeVersion{node6, node8}

        ruby24 := runtimeVersion{runtimeID: "ruby", version: "2.4", httpImage: ruby24Http, pubsubImage: ruby24Pubsub}
        ruby = []runtimeVersion{ruby24}

        netcore2 := runtimeVersion{runtimeID: "netcore", version: "2.0", httpImage: netcore2Http, pubsubImage: ""}
        netcore = []runtimeVersion{netcore2}
```

## 3. Handle the parameters informed by the client

In the function `GetFunctionData()`, include a selection for the `case` structure to define the extention of the file for functions and a variable that will host the file name informed as a parameter for kubeless bringing the code to be executed.

```
func GetFunctionData(runtime, ftype, modName string) (imageName, depName, fileName string, err error) {
        err = nil
        imageName = ""
        depName = ""
        fileName = ""

        runtimeID := regexp.MustCompile("[a-zA-Z]+").FindString(runtime)
        version := regexp.MustCompile("[0-9.]+").FindString(runtime)

        var versionsDef []runtimeVersion
        switch {
        case runtimeID == "python":
                fileName = modName + ".py"
                versionsDef = python
                depName = "requirements.txt"
        case runtimeID == "nodejs":
                fileName = modName + ".js"
                versionsDef = node
                depName = "package.json"
        case runtimeID == "ruby":
                fileName = modName + ".rb"
                versionsDef = ruby
                depName = "Gemfile"
        case runtimeID == "netcore":
                fileName = modName + ".cs"
                versionsDef = netcore
                depName = "project.json"
        default:
                err = errors.New("The given runtime is not valid")
                return
        }
        imageNameEnvVar := ""
```
## 4. Inform an init image as the base stack for your container

In the function `getInitImage()`, include a `case` block to treat if the parameter refers to your new runtime.

```
func getInitImage(runtime string) string {
        switch {
        case strings.Contains(runtime, "python"):
                branch := getBranchFromRuntime(runtime)
                if branch == "2.7" {
                        // TODO: Migrate the image for python 2.7 to an official source (not alpine-based)
                        return "tuna/python-pillow:2.7.11-alpine"
                }
                return "python:" + branch
        case strings.Contains(runtime, "nodejs"):
                return "node:6.10-alpine"
        case strings.Contains(runtime, "ruby"):
                return "bitnami/ruby:2.4"
        case strings.Contains(runtime, "netcore"):
                return "microsoft/aspnetcore-build:2.0"
        default:
                return ""
        }
}
```
As the `return` for the function, inform the name of the base image for your container.

## 5. Set volume mount path for injecting dependencies

In the function `getVolumeMounts()`, specify the path of a mount point that will be available in the container image to inject dependencies.

```
       case strings.Contains(runtime, "netcore"):
                return []v1.VolumeMount{
                        {
                                Name:      "netcorepath",
                                MountPath: "/netcorepath",
                        },
                        {
                                Name:      name,
                                MountPath: "/requirements",
                        },
                }
        default:
                return []v1.VolumeMount{}
        }
}
```

## 6. Update the function to handle requirements for the image

Change the function `updateDeployment()` to include a case to handle your new runtime.

```
        case strings.Contains(runtime, "netcore"):
                dpm.Spec.Template.Spec.Containers[0].Env = append(dpm.Spec.Template.Spec.Containers[0].Env, v1.EnvVar{
                        Name:  "DOTNET_HOME",
                        Value: "/usr/bin/",
                })
                dpm.Spec.Template.Spec.Containers[0].VolumeMounts = append(dpm.Spec.Template.Spec.Containers[0].VolumeMounts, v1.VolumeMount{
                        Name:      "netcorepath",
                        MountPath: "/opt/kubeless/netcorepath",
                })
                dpm.Spec.Template.Spec.Volumes = append(dpm.Spec.Template.Spec.Volumes, v1.Volume{
                        Name: "netcorepath",
                        VolumeSource: v1.VolumeSource{
                                EmptyDir: &v1.EmptyDirVolumeSource{},
                        },
                })
        }
}
```

This function is called if there are requirements to be injected in your runtime or if it is a custom runtime.


After these changes, you need to rebuild the project as described in the Dev Guide.

You will need also to implement a way for your runtime to handle the function code it received by the kuberless client (just in time compilation or something similar).

The runtime should include also a mechanism to receive its parameters and pass them in the right form to your interpreter. 

Stay tuned for future documentations on these additional steps! :-)

